# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("iscsidsc.dll")
prototypes = \
    {
        #
        'GetIScsiVersionInformation': SimTypeFunction([SimTypePointer(SimStruct({"MajorVersion": SimTypeInt(signed=False, label="UInt32"), "MinorVersion": SimTypeInt(signed=False, label="UInt32"), "BuildNumber": SimTypeInt(signed=False, label="UInt32")}, name="ISCSI_VERSION_INFO", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["VersionInfo"]),
        #
        'GetIScsiTargetInformationW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="TARGET_INFORMATION_CLASS"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["TargetName", "DiscoveryMechanism", "InfoClass", "BufferSize", "Buffer"]),
        #
        'GetIScsiTargetInformationA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="TARGET_INFORMATION_CLASS"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["TargetName", "DiscoveryMechanism", "InfoClass", "BufferSize", "Buffer"]),
        #
        'AddIScsiConnectionW': SimTypeFunction([SimTypePointer(SimStruct({"AdapterUnique": SimTypeLongLong(signed=False, label="UInt64"), "AdapterSpecific": SimTypeLongLong(signed=False, label="UInt64")}, name="ISCSI_UNIQUE_SESSION_ID", pack=False, align=None), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"SymbolicName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 256), "Address": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 256), "Socket": SimTypeShort(signed=False, label="UInt16")}, name="ISCSI_TARGET_PORTALW", pack=False, align=None), offset=0), SimTypeLongLong(signed=False, label="UInt64"), SimTypePointer(SimStruct({"Version": SimTypeInt(signed=False, label="UInt32"), "InformationSpecified": SimTypeInt(signed=False, label="UInt32"), "LoginFlags": SimTypeInt(signed=False, label="UInt32"), "AuthType": SimTypeInt(signed=False, label="ISCSI_AUTH_TYPES"), "HeaderDigest": SimTypeInt(signed=False, label="ISCSI_DIGEST_TYPES"), "DataDigest": SimTypeInt(signed=False, label="ISCSI_DIGEST_TYPES"), "MaximumConnections": SimTypeInt(signed=False, label="UInt32"), "DefaultTime2Wait": SimTypeInt(signed=False, label="UInt32"), "DefaultTime2Retain": SimTypeInt(signed=False, label="UInt32"), "UsernameLength": SimTypeInt(signed=False, label="UInt32"), "PasswordLength": SimTypeInt(signed=False, label="UInt32"), "Username": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "Password": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="ISCSI_LOGIN_OPTIONS", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), label="LPArray", offset=0), SimTypePointer(SimStruct({"AdapterUnique": SimTypeLongLong(signed=False, label="UInt64"), "AdapterSpecific": SimTypeLongLong(signed=False, label="UInt64")}, name="ISCSI_UNIQUE_SESSION_ID", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["UniqueSessionId", "Reserved", "InitiatorPortNumber", "TargetPortal", "SecurityFlags", "LoginOptions", "KeySize", "Key", "ConnectionId"]),
        #
        'AddIScsiConnectionA': SimTypeFunction([SimTypePointer(SimStruct({"AdapterUnique": SimTypeLongLong(signed=False, label="UInt64"), "AdapterSpecific": SimTypeLongLong(signed=False, label="UInt64")}, name="ISCSI_UNIQUE_SESSION_ID", pack=False, align=None), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"SymbolicName": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 256), "Address": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 256), "Socket": SimTypeShort(signed=False, label="UInt16")}, name="ISCSI_TARGET_PORTALA", pack=False, align=None), offset=0), SimTypeLongLong(signed=False, label="UInt64"), SimTypePointer(SimStruct({"Version": SimTypeInt(signed=False, label="UInt32"), "InformationSpecified": SimTypeInt(signed=False, label="UInt32"), "LoginFlags": SimTypeInt(signed=False, label="UInt32"), "AuthType": SimTypeInt(signed=False, label="ISCSI_AUTH_TYPES"), "HeaderDigest": SimTypeInt(signed=False, label="ISCSI_DIGEST_TYPES"), "DataDigest": SimTypeInt(signed=False, label="ISCSI_DIGEST_TYPES"), "MaximumConnections": SimTypeInt(signed=False, label="UInt32"), "DefaultTime2Wait": SimTypeInt(signed=False, label="UInt32"), "DefaultTime2Retain": SimTypeInt(signed=False, label="UInt32"), "UsernameLength": SimTypeInt(signed=False, label="UInt32"), "PasswordLength": SimTypeInt(signed=False, label="UInt32"), "Username": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "Password": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="ISCSI_LOGIN_OPTIONS", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), label="LPArray", offset=0), SimTypePointer(SimStruct({"AdapterUnique": SimTypeLongLong(signed=False, label="UInt64"), "AdapterSpecific": SimTypeLongLong(signed=False, label="UInt64")}, name="ISCSI_UNIQUE_SESSION_ID", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["UniqueSessionId", "Reserved", "InitiatorPortNumber", "TargetPortal", "SecurityFlags", "LoginOptions", "KeySize", "Key", "ConnectionId"]),
        #
        'RemoveIScsiConnection': SimTypeFunction([SimTypePointer(SimStruct({"AdapterUnique": SimTypeLongLong(signed=False, label="UInt64"), "AdapterSpecific": SimTypeLongLong(signed=False, label="UInt64")}, name="ISCSI_UNIQUE_SESSION_ID", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"AdapterUnique": SimTypeLongLong(signed=False, label="UInt64"), "AdapterSpecific": SimTypeLongLong(signed=False, label="UInt64")}, name="ISCSI_UNIQUE_SESSION_ID", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["UniqueSessionId", "ConnectionId"]),
        #
        'ReportIScsiTargetsW': SimTypeFunction([SimTypeChar(label="Byte"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ForceUpdate", "BufferSize", "Buffer"]),
        #
        'ReportIScsiTargetsA': SimTypeFunction([SimTypeChar(label="Byte"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), label="LPArray", offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["ForceUpdate", "BufferSize", "Buffer"]),
        #
        'AddIScsiStaticTargetW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeChar(label="Byte"), SimTypePointer(SimStruct({"InitiatorName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 256), "TargetName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 224), "OSDeviceName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 260), "SessionId": SimStruct({"AdapterUnique": SimTypeLongLong(signed=False, label="UInt64"), "AdapterSpecific": SimTypeLongLong(signed=False, label="UInt64")}, name="ISCSI_UNIQUE_SESSION_ID", pack=False, align=None), "OSBusNumber": SimTypeInt(signed=False, label="UInt32"), "OSTargetNumber": SimTypeInt(signed=False, label="UInt32"), "LUNCount": SimTypeInt(signed=False, label="UInt32"), "LUNList": SimTypePointer(SimStruct({"OSLUN": SimTypeInt(signed=False, label="UInt32"), "TargetLUN": SimTypeLongLong(signed=False, label="UInt64")}, name="SCSI_LUN_LIST", pack=False, align=None), offset=0)}, name="ISCSI_TARGET_MAPPINGW", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"Version": SimTypeInt(signed=False, label="UInt32"), "InformationSpecified": SimTypeInt(signed=False, label="UInt32"), "LoginFlags": SimTypeInt(signed=False, label="UInt32"), "AuthType": SimTypeInt(signed=False, label="ISCSI_AUTH_TYPES"), "HeaderDigest": SimTypeInt(signed=False, label="ISCSI_DIGEST_TYPES"), "DataDigest": SimTypeInt(signed=False, label="ISCSI_DIGEST_TYPES"), "MaximumConnections": SimTypeInt(signed=False, label="UInt32"), "DefaultTime2Wait": SimTypeInt(signed=False, label="UInt32"), "DefaultTime2Retain": SimTypeInt(signed=False, label="UInt32"), "UsernameLength": SimTypeInt(signed=False, label="UInt32"), "PasswordLength": SimTypeInt(signed=False, label="UInt32"), "Username": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "Password": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="ISCSI_LOGIN_OPTIONS", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"Count": SimTypeInt(signed=False, label="UInt32"), "Portals": SimTypePointer(SimStruct({"SymbolicName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 256), "Address": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 256), "Socket": SimTypeShort(signed=False, label="UInt16")}, name="ISCSI_TARGET_PORTALW", pack=False, align=None), offset=0)}, name="ISCSI_TARGET_PORTAL_GROUPW", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["TargetName", "TargetAlias", "TargetFlags", "Persist", "Mappings", "LoginOptions", "PortalGroup"]),
        #
        'AddIScsiStaticTargetA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeChar(label="Byte"), SimTypePointer(SimStruct({"InitiatorName": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 256), "TargetName": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 224), "OSDeviceName": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 260), "SessionId": SimStruct({"AdapterUnique": SimTypeLongLong(signed=False, label="UInt64"), "AdapterSpecific": SimTypeLongLong(signed=False, label="UInt64")}, name="ISCSI_UNIQUE_SESSION_ID", pack=False, align=None), "OSBusNumber": SimTypeInt(signed=False, label="UInt32"), "OSTargetNumber": SimTypeInt(signed=False, label="UInt32"), "LUNCount": SimTypeInt(signed=False, label="UInt32"), "LUNList": SimTypePointer(SimStruct({"OSLUN": SimTypeInt(signed=False, label="UInt32"), "TargetLUN": SimTypeLongLong(signed=False, label="UInt64")}, name="SCSI_LUN_LIST", pack=False, align=None), offset=0)}, name="ISCSI_TARGET_MAPPINGA", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"Version": SimTypeInt(signed=False, label="UInt32"), "InformationSpecified": SimTypeInt(signed=False, label="UInt32"), "LoginFlags": SimTypeInt(signed=False, label="UInt32"), "AuthType": SimTypeInt(signed=False, label="ISCSI_AUTH_TYPES"), "HeaderDigest": SimTypeInt(signed=False, label="ISCSI_DIGEST_TYPES"), "DataDigest": SimTypeInt(signed=False, label="ISCSI_DIGEST_TYPES"), "MaximumConnections": SimTypeInt(signed=False, label="UInt32"), "DefaultTime2Wait": SimTypeInt(signed=False, label="UInt32"), "DefaultTime2Retain": SimTypeInt(signed=False, label="UInt32"), "UsernameLength": SimTypeInt(signed=False, label="UInt32"), "PasswordLength": SimTypeInt(signed=False, label="UInt32"), "Username": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "Password": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="ISCSI_LOGIN_OPTIONS", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"Count": SimTypeInt(signed=False, label="UInt32"), "Portals": SimTypePointer(SimStruct({"SymbolicName": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 256), "Address": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 256), "Socket": SimTypeShort(signed=False, label="UInt16")}, name="ISCSI_TARGET_PORTALA", pack=False, align=None), offset=0)}, name="ISCSI_TARGET_PORTAL_GROUPA", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["TargetName", "TargetAlias", "TargetFlags", "Persist", "Mappings", "LoginOptions", "PortalGroup"]),
        #
        'RemoveIScsiStaticTargetW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["TargetName"]),
        #
        'RemoveIScsiStaticTargetA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["TargetName"]),
        #
        'AddIScsiSendTargetPortalW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"Version": SimTypeInt(signed=False, label="UInt32"), "InformationSpecified": SimTypeInt(signed=False, label="UInt32"), "LoginFlags": SimTypeInt(signed=False, label="UInt32"), "AuthType": SimTypeInt(signed=False, label="ISCSI_AUTH_TYPES"), "HeaderDigest": SimTypeInt(signed=False, label="ISCSI_DIGEST_TYPES"), "DataDigest": SimTypeInt(signed=False, label="ISCSI_DIGEST_TYPES"), "MaximumConnections": SimTypeInt(signed=False, label="UInt32"), "DefaultTime2Wait": SimTypeInt(signed=False, label="UInt32"), "DefaultTime2Retain": SimTypeInt(signed=False, label="UInt32"), "UsernameLength": SimTypeInt(signed=False, label="UInt32"), "PasswordLength": SimTypeInt(signed=False, label="UInt32"), "Username": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "Password": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="ISCSI_LOGIN_OPTIONS", pack=False, align=None), offset=0), SimTypeLongLong(signed=False, label="UInt64"), SimTypePointer(SimStruct({"SymbolicName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 256), "Address": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 256), "Socket": SimTypeShort(signed=False, label="UInt16")}, name="ISCSI_TARGET_PORTALW", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["InitiatorInstance", "InitiatorPortNumber", "LoginOptions", "SecurityFlags", "Portal"]),
        #
        'AddIScsiSendTargetPortalA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"Version": SimTypeInt(signed=False, label="UInt32"), "InformationSpecified": SimTypeInt(signed=False, label="UInt32"), "LoginFlags": SimTypeInt(signed=False, label="UInt32"), "AuthType": SimTypeInt(signed=False, label="ISCSI_AUTH_TYPES"), "HeaderDigest": SimTypeInt(signed=False, label="ISCSI_DIGEST_TYPES"), "DataDigest": SimTypeInt(signed=False, label="ISCSI_DIGEST_TYPES"), "MaximumConnections": SimTypeInt(signed=False, label="UInt32"), "DefaultTime2Wait": SimTypeInt(signed=False, label="UInt32"), "DefaultTime2Retain": SimTypeInt(signed=False, label="UInt32"), "UsernameLength": SimTypeInt(signed=False, label="UInt32"), "PasswordLength": SimTypeInt(signed=False, label="UInt32"), "Username": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "Password": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="ISCSI_LOGIN_OPTIONS", pack=False, align=None), offset=0), SimTypeLongLong(signed=False, label="UInt64"), SimTypePointer(SimStruct({"SymbolicName": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 256), "Address": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 256), "Socket": SimTypeShort(signed=False, label="UInt16")}, name="ISCSI_TARGET_PORTALA", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["InitiatorInstance", "InitiatorPortNumber", "LoginOptions", "SecurityFlags", "Portal"]),
        #
        'RemoveIScsiSendTargetPortalW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"SymbolicName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 256), "Address": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 256), "Socket": SimTypeShort(signed=False, label="UInt16")}, name="ISCSI_TARGET_PORTALW", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["InitiatorInstance", "InitiatorPortNumber", "Portal"]),
        #
        'RemoveIScsiSendTargetPortalA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"SymbolicName": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 256), "Address": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 256), "Socket": SimTypeShort(signed=False, label="UInt16")}, name="ISCSI_TARGET_PORTALA", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["InitiatorInstance", "InitiatorPortNumber", "Portal"]),
        #
        'RefreshIScsiSendTargetPortalW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"SymbolicName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 256), "Address": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 256), "Socket": SimTypeShort(signed=False, label="UInt16")}, name="ISCSI_TARGET_PORTALW", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["InitiatorInstance", "InitiatorPortNumber", "Portal"]),
        #
        'RefreshIScsiSendTargetPortalA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"SymbolicName": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 256), "Address": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 256), "Socket": SimTypeShort(signed=False, label="UInt16")}, name="ISCSI_TARGET_PORTALA", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["InitiatorInstance", "InitiatorPortNumber", "Portal"]),
        #
        'ReportIScsiSendTargetPortalsW': SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimStruct({"InitiatorName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 256), "InitiatorPortNumber": SimTypeInt(signed=False, label="UInt32"), "SymbolicName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 256), "Address": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 256), "Socket": SimTypeShort(signed=False, label="UInt16")}, name="ISCSI_TARGET_PORTAL_INFOW", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["PortalCount", "PortalInfo"]),
        #
        'ReportIScsiSendTargetPortalsA': SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimStruct({"InitiatorName": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 256), "InitiatorPortNumber": SimTypeInt(signed=False, label="UInt32"), "SymbolicName": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 256), "Address": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 256), "Socket": SimTypeShort(signed=False, label="UInt16")}, name="ISCSI_TARGET_PORTAL_INFOA", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["PortalCount", "PortalInfo"]),
        #
        'ReportIScsiSendTargetPortalsExW': SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimStruct({"InitiatorName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 256), "InitiatorPortNumber": SimTypeInt(signed=False, label="UInt32"), "SymbolicName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 256), "Address": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 256), "Socket": SimTypeShort(signed=False, label="UInt16"), "SecurityFlags": SimTypeLongLong(signed=False, label="UInt64"), "LoginOptions": SimStruct({"Version": SimTypeInt(signed=False, label="UInt32"), "InformationSpecified": SimTypeInt(signed=False, label="UInt32"), "LoginFlags": SimTypeInt(signed=False, label="UInt32"), "AuthType": SimTypeInt(signed=False, label="ISCSI_AUTH_TYPES"), "HeaderDigest": SimTypeInt(signed=False, label="ISCSI_DIGEST_TYPES"), "DataDigest": SimTypeInt(signed=False, label="ISCSI_DIGEST_TYPES"), "MaximumConnections": SimTypeInt(signed=False, label="UInt32"), "DefaultTime2Wait": SimTypeInt(signed=False, label="UInt32"), "DefaultTime2Retain": SimTypeInt(signed=False, label="UInt32"), "UsernameLength": SimTypeInt(signed=False, label="UInt32"), "PasswordLength": SimTypeInt(signed=False, label="UInt32"), "Username": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "Password": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="ISCSI_LOGIN_OPTIONS", pack=False, align=None)}, name="ISCSI_TARGET_PORTAL_INFO_EXW", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["PortalCount", "PortalInfoSize", "PortalInfo"]),
        #
        'ReportIScsiSendTargetPortalsExA': SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimStruct({"InitiatorName": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 256), "InitiatorPortNumber": SimTypeInt(signed=False, label="UInt32"), "SymbolicName": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 256), "Address": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 256), "Socket": SimTypeShort(signed=False, label="UInt16"), "SecurityFlags": SimTypeLongLong(signed=False, label="UInt64"), "LoginOptions": SimStruct({"Version": SimTypeInt(signed=False, label="UInt32"), "InformationSpecified": SimTypeInt(signed=False, label="UInt32"), "LoginFlags": SimTypeInt(signed=False, label="UInt32"), "AuthType": SimTypeInt(signed=False, label="ISCSI_AUTH_TYPES"), "HeaderDigest": SimTypeInt(signed=False, label="ISCSI_DIGEST_TYPES"), "DataDigest": SimTypeInt(signed=False, label="ISCSI_DIGEST_TYPES"), "MaximumConnections": SimTypeInt(signed=False, label="UInt32"), "DefaultTime2Wait": SimTypeInt(signed=False, label="UInt32"), "DefaultTime2Retain": SimTypeInt(signed=False, label="UInt32"), "UsernameLength": SimTypeInt(signed=False, label="UInt32"), "PasswordLength": SimTypeInt(signed=False, label="UInt32"), "Username": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "Password": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="ISCSI_LOGIN_OPTIONS", pack=False, align=None)}, name="ISCSI_TARGET_PORTAL_INFO_EXA", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["PortalCount", "PortalInfoSize", "PortalInfo"]),
        #
        'LoginIScsiTargetW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeChar(label="Byte"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"SymbolicName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 256), "Address": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 256), "Socket": SimTypeShort(signed=False, label="UInt16")}, name="ISCSI_TARGET_PORTALW", pack=False, align=None), offset=0), SimTypeLongLong(signed=False, label="UInt64"), SimTypePointer(SimStruct({"InitiatorName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 256), "TargetName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 224), "OSDeviceName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 260), "SessionId": SimStruct({"AdapterUnique": SimTypeLongLong(signed=False, label="UInt64"), "AdapterSpecific": SimTypeLongLong(signed=False, label="UInt64")}, name="ISCSI_UNIQUE_SESSION_ID", pack=False, align=None), "OSBusNumber": SimTypeInt(signed=False, label="UInt32"), "OSTargetNumber": SimTypeInt(signed=False, label="UInt32"), "LUNCount": SimTypeInt(signed=False, label="UInt32"), "LUNList": SimTypePointer(SimStruct({"OSLUN": SimTypeInt(signed=False, label="UInt32"), "TargetLUN": SimTypeLongLong(signed=False, label="UInt64")}, name="SCSI_LUN_LIST", pack=False, align=None), offset=0)}, name="ISCSI_TARGET_MAPPINGW", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"Version": SimTypeInt(signed=False, label="UInt32"), "InformationSpecified": SimTypeInt(signed=False, label="UInt32"), "LoginFlags": SimTypeInt(signed=False, label="UInt32"), "AuthType": SimTypeInt(signed=False, label="ISCSI_AUTH_TYPES"), "HeaderDigest": SimTypeInt(signed=False, label="ISCSI_DIGEST_TYPES"), "DataDigest": SimTypeInt(signed=False, label="ISCSI_DIGEST_TYPES"), "MaximumConnections": SimTypeInt(signed=False, label="UInt32"), "DefaultTime2Wait": SimTypeInt(signed=False, label="UInt32"), "DefaultTime2Retain": SimTypeInt(signed=False, label="UInt32"), "UsernameLength": SimTypeInt(signed=False, label="UInt32"), "PasswordLength": SimTypeInt(signed=False, label="UInt32"), "Username": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "Password": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="ISCSI_LOGIN_OPTIONS", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), label="LPArray", offset=0), SimTypeChar(label="Byte"), SimTypePointer(SimStruct({"AdapterUnique": SimTypeLongLong(signed=False, label="UInt64"), "AdapterSpecific": SimTypeLongLong(signed=False, label="UInt64")}, name="ISCSI_UNIQUE_SESSION_ID", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"AdapterUnique": SimTypeLongLong(signed=False, label="UInt64"), "AdapterSpecific": SimTypeLongLong(signed=False, label="UInt64")}, name="ISCSI_UNIQUE_SESSION_ID", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["TargetName", "IsInformationalSession", "InitiatorInstance", "InitiatorPortNumber", "TargetPortal", "SecurityFlags", "Mappings", "LoginOptions", "KeySize", "Key", "IsPersistent", "UniqueSessionId", "UniqueConnectionId"]),
        #
        'LoginIScsiTargetA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeChar(label="Byte"), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"SymbolicName": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 256), "Address": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 256), "Socket": SimTypeShort(signed=False, label="UInt16")}, name="ISCSI_TARGET_PORTALA", pack=False, align=None), offset=0), SimTypeLongLong(signed=False, label="UInt64"), SimTypePointer(SimStruct({"InitiatorName": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 256), "TargetName": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 224), "OSDeviceName": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 260), "SessionId": SimStruct({"AdapterUnique": SimTypeLongLong(signed=False, label="UInt64"), "AdapterSpecific": SimTypeLongLong(signed=False, label="UInt64")}, name="ISCSI_UNIQUE_SESSION_ID", pack=False, align=None), "OSBusNumber": SimTypeInt(signed=False, label="UInt32"), "OSTargetNumber": SimTypeInt(signed=False, label="UInt32"), "LUNCount": SimTypeInt(signed=False, label="UInt32"), "LUNList": SimTypePointer(SimStruct({"OSLUN": SimTypeInt(signed=False, label="UInt32"), "TargetLUN": SimTypeLongLong(signed=False, label="UInt64")}, name="SCSI_LUN_LIST", pack=False, align=None), offset=0)}, name="ISCSI_TARGET_MAPPINGA", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"Version": SimTypeInt(signed=False, label="UInt32"), "InformationSpecified": SimTypeInt(signed=False, label="UInt32"), "LoginFlags": SimTypeInt(signed=False, label="UInt32"), "AuthType": SimTypeInt(signed=False, label="ISCSI_AUTH_TYPES"), "HeaderDigest": SimTypeInt(signed=False, label="ISCSI_DIGEST_TYPES"), "DataDigest": SimTypeInt(signed=False, label="ISCSI_DIGEST_TYPES"), "MaximumConnections": SimTypeInt(signed=False, label="UInt32"), "DefaultTime2Wait": SimTypeInt(signed=False, label="UInt32"), "DefaultTime2Retain": SimTypeInt(signed=False, label="UInt32"), "UsernameLength": SimTypeInt(signed=False, label="UInt32"), "PasswordLength": SimTypeInt(signed=False, label="UInt32"), "Username": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "Password": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="ISCSI_LOGIN_OPTIONS", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), label="LPArray", offset=0), SimTypeChar(label="Byte"), SimTypePointer(SimStruct({"AdapterUnique": SimTypeLongLong(signed=False, label="UInt64"), "AdapterSpecific": SimTypeLongLong(signed=False, label="UInt64")}, name="ISCSI_UNIQUE_SESSION_ID", pack=False, align=None), offset=0), SimTypePointer(SimStruct({"AdapterUnique": SimTypeLongLong(signed=False, label="UInt64"), "AdapterSpecific": SimTypeLongLong(signed=False, label="UInt64")}, name="ISCSI_UNIQUE_SESSION_ID", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["TargetName", "IsInformationalSession", "InitiatorInstance", "InitiatorPortNumber", "TargetPortal", "SecurityFlags", "Mappings", "LoginOptions", "KeySize", "Key", "IsPersistent", "UniqueSessionId", "UniqueConnectionId"]),
        #
        'ReportIScsiPersistentLoginsW': SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimStruct({"TargetName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 224), "IsInformationalSession": SimTypeChar(label="Byte"), "InitiatorInstance": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 256), "InitiatorPortNumber": SimTypeInt(signed=False, label="UInt32"), "TargetPortal": SimStruct({"SymbolicName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 256), "Address": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 256), "Socket": SimTypeShort(signed=False, label="UInt16")}, name="ISCSI_TARGET_PORTALW", pack=False, align=None), "SecurityFlags": SimTypeLongLong(signed=False, label="UInt64"), "Mappings": SimTypePointer(SimStruct({"InitiatorName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 256), "TargetName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 224), "OSDeviceName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 260), "SessionId": SimStruct({"AdapterUnique": SimTypeLongLong(signed=False, label="UInt64"), "AdapterSpecific": SimTypeLongLong(signed=False, label="UInt64")}, name="ISCSI_UNIQUE_SESSION_ID", pack=False, align=None), "OSBusNumber": SimTypeInt(signed=False, label="UInt32"), "OSTargetNumber": SimTypeInt(signed=False, label="UInt32"), "LUNCount": SimTypeInt(signed=False, label="UInt32"), "LUNList": SimTypePointer(SimStruct({"OSLUN": SimTypeInt(signed=False, label="UInt32"), "TargetLUN": SimTypeLongLong(signed=False, label="UInt64")}, name="SCSI_LUN_LIST", pack=False, align=None), offset=0)}, name="ISCSI_TARGET_MAPPINGW", pack=False, align=None), offset=0), "LoginOptions": SimStruct({"Version": SimTypeInt(signed=False, label="UInt32"), "InformationSpecified": SimTypeInt(signed=False, label="UInt32"), "LoginFlags": SimTypeInt(signed=False, label="UInt32"), "AuthType": SimTypeInt(signed=False, label="ISCSI_AUTH_TYPES"), "HeaderDigest": SimTypeInt(signed=False, label="ISCSI_DIGEST_TYPES"), "DataDigest": SimTypeInt(signed=False, label="ISCSI_DIGEST_TYPES"), "MaximumConnections": SimTypeInt(signed=False, label="UInt32"), "DefaultTime2Wait": SimTypeInt(signed=False, label="UInt32"), "DefaultTime2Retain": SimTypeInt(signed=False, label="UInt32"), "UsernameLength": SimTypeInt(signed=False, label="UInt32"), "PasswordLength": SimTypeInt(signed=False, label="UInt32"), "Username": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "Password": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="ISCSI_LOGIN_OPTIONS", pack=False, align=None)}, name="PERSISTENT_ISCSI_LOGIN_INFOW", pack=False, align=None), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["Count", "PersistentLoginInfo", "BufferSizeInBytes"]),
        #
        'ReportIScsiPersistentLoginsA': SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimStruct({"TargetName": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 224), "IsInformationalSession": SimTypeChar(label="Byte"), "InitiatorInstance": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 256), "InitiatorPortNumber": SimTypeInt(signed=False, label="UInt32"), "TargetPortal": SimStruct({"SymbolicName": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 256), "Address": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 256), "Socket": SimTypeShort(signed=False, label="UInt16")}, name="ISCSI_TARGET_PORTALA", pack=False, align=None), "SecurityFlags": SimTypeLongLong(signed=False, label="UInt64"), "Mappings": SimTypePointer(SimStruct({"InitiatorName": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 256), "TargetName": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 224), "OSDeviceName": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 260), "SessionId": SimStruct({"AdapterUnique": SimTypeLongLong(signed=False, label="UInt64"), "AdapterSpecific": SimTypeLongLong(signed=False, label="UInt64")}, name="ISCSI_UNIQUE_SESSION_ID", pack=False, align=None), "OSBusNumber": SimTypeInt(signed=False, label="UInt32"), "OSTargetNumber": SimTypeInt(signed=False, label="UInt32"), "LUNCount": SimTypeInt(signed=False, label="UInt32"), "LUNList": SimTypePointer(SimStruct({"OSLUN": SimTypeInt(signed=False, label="UInt32"), "TargetLUN": SimTypeLongLong(signed=False, label="UInt64")}, name="SCSI_LUN_LIST", pack=False, align=None), offset=0)}, name="ISCSI_TARGET_MAPPINGA", pack=False, align=None), offset=0), "LoginOptions": SimStruct({"Version": SimTypeInt(signed=False, label="UInt32"), "InformationSpecified": SimTypeInt(signed=False, label="UInt32"), "LoginFlags": SimTypeInt(signed=False, label="UInt32"), "AuthType": SimTypeInt(signed=False, label="ISCSI_AUTH_TYPES"), "HeaderDigest": SimTypeInt(signed=False, label="ISCSI_DIGEST_TYPES"), "DataDigest": SimTypeInt(signed=False, label="ISCSI_DIGEST_TYPES"), "MaximumConnections": SimTypeInt(signed=False, label="UInt32"), "DefaultTime2Wait": SimTypeInt(signed=False, label="UInt32"), "DefaultTime2Retain": SimTypeInt(signed=False, label="UInt32"), "UsernameLength": SimTypeInt(signed=False, label="UInt32"), "PasswordLength": SimTypeInt(signed=False, label="UInt32"), "Username": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "Password": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="ISCSI_LOGIN_OPTIONS", pack=False, align=None)}, name="PERSISTENT_ISCSI_LOGIN_INFOA", pack=False, align=None), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["Count", "PersistentLoginInfo", "BufferSizeInBytes"]),
        #
        'LogoutIScsiTarget': SimTypeFunction([SimTypePointer(SimStruct({"AdapterUnique": SimTypeLongLong(signed=False, label="UInt64"), "AdapterSpecific": SimTypeLongLong(signed=False, label="UInt64")}, name="ISCSI_UNIQUE_SESSION_ID", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["UniqueSessionId"]),
        #
        'RemoveIScsiPersistentTargetW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimStruct({"SymbolicName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 256), "Address": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 256), "Socket": SimTypeShort(signed=False, label="UInt16")}, name="ISCSI_TARGET_PORTALW", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["InitiatorInstance", "InitiatorPortNumber", "TargetName", "Portal"]),
        #
        'RemoveIScsiPersistentTargetA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimStruct({"SymbolicName": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 256), "Address": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 256), "Socket": SimTypeShort(signed=False, label="UInt16")}, name="ISCSI_TARGET_PORTALA", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["InitiatorInstance", "InitiatorPortNumber", "TargetName", "Portal"]),
        #
        'SendScsiInquiry': SimTypeFunction([SimTypePointer(SimStruct({"AdapterUnique": SimTypeLongLong(signed=False, label="UInt64"), "AdapterSpecific": SimTypeLongLong(signed=False, label="UInt64")}, name="ISCSI_UNIQUE_SESSION_ID", pack=False, align=None), offset=0), SimTypeLongLong(signed=False, label="UInt64"), SimTypeChar(label="Byte"), SimTypeChar(label="Byte"), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["UniqueSessionId", "Lun", "EvpdCmddt", "PageCode", "ScsiStatus", "ResponseSize", "ResponseBuffer", "SenseSize", "SenseBuffer"]),
        #
        'SendScsiReadCapacity': SimTypeFunction([SimTypePointer(SimStruct({"AdapterUnique": SimTypeLongLong(signed=False, label="UInt64"), "AdapterSpecific": SimTypeLongLong(signed=False, label="UInt64")}, name="ISCSI_UNIQUE_SESSION_ID", pack=False, align=None), offset=0), SimTypeLongLong(signed=False, label="UInt64"), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["UniqueSessionId", "Lun", "ScsiStatus", "ResponseSize", "ResponseBuffer", "SenseSize", "SenseBuffer"]),
        #
        'SendScsiReportLuns': SimTypeFunction([SimTypePointer(SimStruct({"AdapterUnique": SimTypeLongLong(signed=False, label="UInt64"), "AdapterSpecific": SimTypeLongLong(signed=False, label="UInt64")}, name="ISCSI_UNIQUE_SESSION_ID", pack=False, align=None), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["UniqueSessionId", "ScsiStatus", "ResponseSize", "ResponseBuffer", "SenseSize", "SenseBuffer"]),
        #
        'ReportIScsiInitiatorListW': SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["BufferSize", "Buffer"]),
        #
        'ReportIScsiInitiatorListA': SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), label="LPArray", offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["BufferSize", "Buffer"]),
        #
        'ReportActiveIScsiTargetMappingsW': SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimStruct({"InitiatorName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 256), "TargetName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 224), "OSDeviceName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 260), "SessionId": SimStruct({"AdapterUnique": SimTypeLongLong(signed=False, label="UInt64"), "AdapterSpecific": SimTypeLongLong(signed=False, label="UInt64")}, name="ISCSI_UNIQUE_SESSION_ID", pack=False, align=None), "OSBusNumber": SimTypeInt(signed=False, label="UInt32"), "OSTargetNumber": SimTypeInt(signed=False, label="UInt32"), "LUNCount": SimTypeInt(signed=False, label="UInt32"), "LUNList": SimTypePointer(SimStruct({"OSLUN": SimTypeInt(signed=False, label="UInt32"), "TargetLUN": SimTypeLongLong(signed=False, label="UInt64")}, name="SCSI_LUN_LIST", pack=False, align=None), offset=0)}, name="ISCSI_TARGET_MAPPINGW", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["BufferSize", "MappingCount", "Mappings"]),
        #
        'ReportActiveIScsiTargetMappingsA': SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimStruct({"InitiatorName": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 256), "TargetName": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 224), "OSDeviceName": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 260), "SessionId": SimStruct({"AdapterUnique": SimTypeLongLong(signed=False, label="UInt64"), "AdapterSpecific": SimTypeLongLong(signed=False, label="UInt64")}, name="ISCSI_UNIQUE_SESSION_ID", pack=False, align=None), "OSBusNumber": SimTypeInt(signed=False, label="UInt32"), "OSTargetNumber": SimTypeInt(signed=False, label="UInt32"), "LUNCount": SimTypeInt(signed=False, label="UInt32"), "LUNList": SimTypePointer(SimStruct({"OSLUN": SimTypeInt(signed=False, label="UInt32"), "TargetLUN": SimTypeLongLong(signed=False, label="UInt64")}, name="SCSI_LUN_LIST", pack=False, align=None), offset=0)}, name="ISCSI_TARGET_MAPPINGA", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["BufferSize", "MappingCount", "Mappings"]),
        #
        'SetIScsiTunnelModeOuterAddressW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeChar(label="Byte")], SimTypeInt(signed=False, label="UInt32"), arg_names=["InitiatorName", "InitiatorPortNumber", "DestinationAddress", "OuterModeAddress", "Persist"]),
        #
        'SetIScsiTunnelModeOuterAddressA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeChar(label="Byte")], SimTypeInt(signed=False, label="UInt32"), arg_names=["InitiatorName", "InitiatorPortNumber", "DestinationAddress", "OuterModeAddress", "Persist"]),
        #
        'SetIScsiIKEInfoW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"AuthMethod": SimTypeInt(signed=False, label="IKE_AUTHENTICATION_METHOD"), "Anonymous": SimUnion({"PsKey": SimStruct({"SecurityFlags": SimTypeLongLong(signed=False, label="UInt64"), "IdType": SimTypeChar(label="Byte"), "IdLengthInBytes": SimTypeInt(signed=False, label="UInt32"), "Id": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "KeyLengthInBytes": SimTypeInt(signed=False, label="UInt32"), "Key": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="IKE_AUTHENTICATION_PRESHARED_KEY", pack=False, align=None)}, name="<anon>", label="None")}, name="IKE_AUTHENTICATION_INFORMATION", pack=False, align=None), offset=0), SimTypeChar(label="Byte")], SimTypeInt(signed=False, label="UInt32"), arg_names=["InitiatorName", "InitiatorPortNumber", "AuthInfo", "Persist"]),
        #
        'SetIScsiIKEInfoA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"AuthMethod": SimTypeInt(signed=False, label="IKE_AUTHENTICATION_METHOD"), "Anonymous": SimUnion({"PsKey": SimStruct({"SecurityFlags": SimTypeLongLong(signed=False, label="UInt64"), "IdType": SimTypeChar(label="Byte"), "IdLengthInBytes": SimTypeInt(signed=False, label="UInt32"), "Id": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "KeyLengthInBytes": SimTypeInt(signed=False, label="UInt32"), "Key": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="IKE_AUTHENTICATION_PRESHARED_KEY", pack=False, align=None)}, name="<anon>", label="None")}, name="IKE_AUTHENTICATION_INFORMATION", pack=False, align=None), offset=0), SimTypeChar(label="Byte")], SimTypeInt(signed=False, label="UInt32"), arg_names=["InitiatorName", "InitiatorPortNumber", "AuthInfo", "Persist"]),
        #
        'GetIScsiIKEInfoW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimStruct({"AuthMethod": SimTypeInt(signed=False, label="IKE_AUTHENTICATION_METHOD"), "Anonymous": SimUnion({"PsKey": SimStruct({"SecurityFlags": SimTypeLongLong(signed=False, label="UInt64"), "IdType": SimTypeChar(label="Byte"), "IdLengthInBytes": SimTypeInt(signed=False, label="UInt32"), "Id": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "KeyLengthInBytes": SimTypeInt(signed=False, label="UInt32"), "Key": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="IKE_AUTHENTICATION_PRESHARED_KEY", pack=False, align=None)}, name="<anon>", label="None")}, name="IKE_AUTHENTICATION_INFORMATION", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["InitiatorName", "InitiatorPortNumber", "Reserved", "AuthInfo"]),
        #
        'GetIScsiIKEInfoA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimStruct({"AuthMethod": SimTypeInt(signed=False, label="IKE_AUTHENTICATION_METHOD"), "Anonymous": SimUnion({"PsKey": SimStruct({"SecurityFlags": SimTypeLongLong(signed=False, label="UInt64"), "IdType": SimTypeChar(label="Byte"), "IdLengthInBytes": SimTypeInt(signed=False, label="UInt32"), "Id": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "KeyLengthInBytes": SimTypeInt(signed=False, label="UInt32"), "Key": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="IKE_AUTHENTICATION_PRESHARED_KEY", pack=False, align=None)}, name="<anon>", label="None")}, name="IKE_AUTHENTICATION_INFORMATION", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["InitiatorName", "InitiatorPortNumber", "Reserved", "AuthInfo"]),
        #
        'SetIScsiGroupPresharedKey': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypeChar(label="Byte")], SimTypeInt(signed=False, label="UInt32"), arg_names=["KeyLength", "Key", "Persist"]),
        #
        'SetIScsiInitiatorCHAPSharedSecret': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["SharedSecretLength", "SharedSecret"]),
        #
        'SetIScsiInitiatorRADIUSSharedSecret': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["SharedSecretLength", "SharedSecret"]),
        #
        'SetIScsiInitiatorNodeNameW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["InitiatorNodeName"]),
        #
        'SetIScsiInitiatorNodeNameA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["InitiatorNodeName"]),
        #
        'GetIScsiInitiatorNodeNameW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["InitiatorNodeName"]),
        #
        'GetIScsiInitiatorNodeNameA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["InitiatorNodeName"]),
        #
        'AddISNSServerW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["Address"]),
        #
        'AddISNSServerA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["Address"]),
        #
        'RemoveISNSServerW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["Address"]),
        #
        'RemoveISNSServerA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["Address"]),
        #
        'RefreshISNSServerW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["Address"]),
        #
        'RefreshISNSServerA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["Address"]),
        #
        'ReportISNSServerListW': SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["BufferSizeInChar", "Buffer"]),
        #
        'ReportISNSServerListA': SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), label="LPArray", offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["BufferSizeInChar", "Buffer"]),
        #
        'GetIScsiSessionListW': SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimStruct({"SessionId": SimStruct({"AdapterUnique": SimTypeLongLong(signed=False, label="UInt64"), "AdapterSpecific": SimTypeLongLong(signed=False, label="UInt64")}, name="ISCSI_UNIQUE_SESSION_ID", pack=False, align=None), "InitiatorName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "TargetNodeName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "TargetName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "ISID": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 6), "TSID": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "ConnectionCount": SimTypeInt(signed=False, label="UInt32"), "Connections": SimTypePointer(SimStruct({"ConnectionId": SimStruct({"AdapterUnique": SimTypeLongLong(signed=False, label="UInt64"), "AdapterSpecific": SimTypeLongLong(signed=False, label="UInt64")}, name="ISCSI_UNIQUE_SESSION_ID", pack=False, align=None), "InitiatorAddress": SimTypePointer(SimTypeChar(label="Char"), offset=0), "TargetAddress": SimTypePointer(SimTypeChar(label="Char"), offset=0), "InitiatorSocket": SimTypeShort(signed=False, label="UInt16"), "TargetSocket": SimTypeShort(signed=False, label="UInt16"), "CID": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2)}, name="ISCSI_CONNECTION_INFOW", pack=False, align=None), offset=0)}, name="ISCSI_SESSION_INFOW", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["BufferSize", "SessionCount", "SessionInfo"]),
        #
        'GetIScsiSessionListA': SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimStruct({"SessionId": SimStruct({"AdapterUnique": SimTypeLongLong(signed=False, label="UInt64"), "AdapterSpecific": SimTypeLongLong(signed=False, label="UInt64")}, name="ISCSI_UNIQUE_SESSION_ID", pack=False, align=None), "InitiatorName": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "TargetNodeName": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "TargetName": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "ISID": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 6), "TSID": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2), "ConnectionCount": SimTypeInt(signed=False, label="UInt32"), "Connections": SimTypePointer(SimStruct({"ConnectionId": SimStruct({"AdapterUnique": SimTypeLongLong(signed=False, label="UInt64"), "AdapterSpecific": SimTypeLongLong(signed=False, label="UInt64")}, name="ISCSI_UNIQUE_SESSION_ID", pack=False, align=None), "InitiatorAddress": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "TargetAddress": SimTypePointer(SimTypeChar(label="Byte"), offset=0), "InitiatorSocket": SimTypeShort(signed=False, label="UInt16"), "TargetSocket": SimTypeShort(signed=False, label="UInt16"), "CID": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 2)}, name="ISCSI_CONNECTION_INFOA", pack=False, align=None), offset=0)}, name="ISCSI_SESSION_INFOA", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["BufferSize", "SessionCount", "SessionInfo"]),
        #
        'GetIScsiSessionListEx': SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimStruct({"SessionId": SimStruct({"AdapterUnique": SimTypeLongLong(signed=False, label="UInt64"), "AdapterSpecific": SimTypeLongLong(signed=False, label="UInt64")}, name="ISCSI_UNIQUE_SESSION_ID", pack=False, align=None), "InitialR2t": SimTypeChar(label="Byte"), "ImmediateData": SimTypeChar(label="Byte"), "Type": SimTypeChar(label="Byte"), "DataSequenceInOrder": SimTypeChar(label="Byte"), "DataPduInOrder": SimTypeChar(label="Byte"), "ErrorRecoveryLevel": SimTypeChar(label="Byte"), "MaxOutstandingR2t": SimTypeInt(signed=False, label="UInt32"), "FirstBurstLength": SimTypeInt(signed=False, label="UInt32"), "MaxBurstLength": SimTypeInt(signed=False, label="UInt32"), "MaximumConnections": SimTypeInt(signed=False, label="UInt32"), "ConnectionCount": SimTypeInt(signed=False, label="UInt32"), "Connections": SimTypePointer(SimStruct({"ConnectionId": SimStruct({"AdapterUnique": SimTypeLongLong(signed=False, label="UInt64"), "AdapterSpecific": SimTypeLongLong(signed=False, label="UInt64")}, name="ISCSI_UNIQUE_SESSION_ID", pack=False, align=None), "State": SimTypeChar(label="Byte"), "Protocol": SimTypeChar(label="Byte"), "HeaderDigest": SimTypeChar(label="Byte"), "DataDigest": SimTypeChar(label="Byte"), "MaxRecvDataSegmentLength": SimTypeInt(signed=False, label="UInt32"), "AuthType": SimTypeInt(signed=False, label="ISCSI_AUTH_TYPES"), "EstimatedThroughput": SimTypeLongLong(signed=False, label="UInt64"), "MaxDatagramSize": SimTypeInt(signed=False, label="UInt32")}, name="ISCSI_CONNECTION_INFO_EX", pack=False, align=None), offset=0)}, name="ISCSI_SESSION_INFO_EX", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["BufferSize", "SessionCountPtr", "SessionInfo"]),
        #
        'GetDevicesForIScsiSessionW': SimTypeFunction([SimTypePointer(SimStruct({"AdapterUnique": SimTypeLongLong(signed=False, label="UInt64"), "AdapterSpecific": SimTypeLongLong(signed=False, label="UInt64")}, name="ISCSI_UNIQUE_SESSION_ID", pack=False, align=None), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimStruct({"InitiatorName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 256), "TargetName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 224), "ScsiAddress": SimStruct({"Length": SimTypeInt(signed=False, label="UInt32"), "PortNumber": SimTypeChar(label="Byte"), "PathId": SimTypeChar(label="Byte"), "TargetId": SimTypeChar(label="Byte"), "Lun": SimTypeChar(label="Byte")}, name="SCSI_ADDRESS", pack=False, align=None), "DeviceInterfaceType": SimTypeBottom(label="Guid"), "DeviceInterfaceName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 260), "LegacyName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 260), "StorageDeviceNumber": SimTypeBottom(label="STORAGE_DEVICE_NUMBER"), "DeviceInstance": SimTypeInt(signed=False, label="UInt32")}, name="ISCSI_DEVICE_ON_SESSIONW", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["UniqueSessionId", "DeviceCount", "Devices"]),
        #
        'GetDevicesForIScsiSessionA': SimTypeFunction([SimTypePointer(SimStruct({"AdapterUnique": SimTypeLongLong(signed=False, label="UInt64"), "AdapterSpecific": SimTypeLongLong(signed=False, label="UInt64")}, name="ISCSI_UNIQUE_SESSION_ID", pack=False, align=None), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimStruct({"InitiatorName": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 256), "TargetName": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 224), "ScsiAddress": SimStruct({"Length": SimTypeInt(signed=False, label="UInt32"), "PortNumber": SimTypeChar(label="Byte"), "PathId": SimTypeChar(label="Byte"), "TargetId": SimTypeChar(label="Byte"), "Lun": SimTypeChar(label="Byte")}, name="SCSI_ADDRESS", pack=False, align=None), "DeviceInterfaceType": SimTypeBottom(label="Guid"), "DeviceInterfaceName": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 260), "LegacyName": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 260), "StorageDeviceNumber": SimTypeBottom(label="STORAGE_DEVICE_NUMBER"), "DeviceInstance": SimTypeInt(signed=False, label="UInt32")}, name="ISCSI_DEVICE_ON_SESSIONA", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["UniqueSessionId", "DeviceCount", "Devices"]),
        #
        'SetupPersistentIScsiVolumes': SimTypeFunction([], SimTypeInt(signed=False, label="UInt32")),
        #
        'SetupPersistentIScsiDevices': SimTypeFunction([], SimTypeInt(signed=False, label="UInt32")),
        #
        'AddPersistentIScsiDeviceW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["DevicePath"]),
        #
        'AddPersistentIScsiDeviceA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["DevicePath"]),
        #
        'RemovePersistentIScsiDeviceW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["DevicePath"]),
        #
        'RemovePersistentIScsiDeviceA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["DevicePath"]),
        #
        'ClearPersistentIScsiDevices': SimTypeFunction([], SimTypeInt(signed=False, label="UInt32")),
        #
        'ReportPersistentIScsiDevicesW': SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["BufferSizeInChar", "Buffer"]),
        #
        'ReportPersistentIScsiDevicesA': SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), label="LPArray", offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["BufferSizeInChar", "Buffer"]),
        #
        'ReportIScsiTargetPortalsW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeShort(signed=False, label="UInt16"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimStruct({"SymbolicName": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 256), "Address": SimTypeFixedSizeArray(SimTypeChar(label="Char"), 256), "Socket": SimTypeShort(signed=False, label="UInt16")}, name="ISCSI_TARGET_PORTALW", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["InitiatorName", "TargetName", "TargetPortalTag", "ElementCount", "Portals"]),
        #
        'ReportIScsiTargetPortalsA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), offset=0), SimTypePointer(SimTypeShort(signed=False, label="UInt16"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimStruct({"SymbolicName": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 256), "Address": SimTypeFixedSizeArray(SimTypeBottom(label="CHAR"), 256), "Socket": SimTypeShort(signed=False, label="UInt16")}, name="ISCSI_TARGET_PORTALA", pack=False, align=None), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["InitiatorName", "TargetName", "TargetPortalTag", "ElementCount", "Portals"]),
        #
        'AddRadiusServerW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["Address"]),
        #
        'AddRadiusServerA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["Address"]),
        #
        'RemoveRadiusServerW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["Address"]),
        #
        'RemoveRadiusServerA': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["Address"]),
        #
        'ReportRadiusServerListW': SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Char"), label="LPArray", offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["BufferSizeInChar", "Buffer"]),
        #
        'ReportRadiusServerListA': SimTypeFunction([SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypeChar(label="Byte"), label="LPArray", offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["BufferSizeInChar", "Buffer"]),
    }

lib.set_prototypes(prototypes)



import angr


class GetKeyboardType(angr.SimProcedure):
    def run(self, param):
        # return the values present at time of author's testing
        if self.state.solver.is_true(param == 0):
            return 4
        if self.state.solver.is_true(param == 1):
            return 0
        if self.state.solver.is_true(param == 2):
            return 12
        return 0



import copy
import functools
import logging

import archinfo

from ..errors import SimIRSBError, SimIRSBNoDecodeError, SimValueError
from .engine import SuccessorsMixin
from .vex.heavy.heavy import VEXEarlyExit
from .. import sim_options as o
from ..misc.ux import once
from ..state_plugins.inspect import BP_AFTER, BP_BEFORE
from ..state_plugins.unicorn_engine import STOP, _UC_NATIVE, unicorn as uc_module
from ..utils.constants import DEFAULT_STATEMENT

# pylint: disable=arguments-differ

l = logging.getLogger(name=__name__)


class SimEngineUnicorn(SuccessorsMixin):
    """
    Concrete execution in the Unicorn Engine, a fork of qemu.

    Responds to the following parameters in the step stack:

    - step:                How many basic blocks we want to execute
    - extra_stop_points:   A collection of addresses at which execution should halt
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Cache of details of basic blocks containing statements that need to re-executed
        self._block_details_cache = {}
        # Addresses of basic blocks which native interface will not execute
        self._stop_block_addrs_cache = set()
        # Stop reasons to track and not switch to native interface for those basic blocks
        self._stop_reasons_to_track = STOP.unsupported_reasons | {
            STOP.STOP_STOPPOINT,
            STOP.STOP_ERROR,
            STOP.STOP_NODECODE,
            STOP.STOP_SYSCALL,
            STOP.STOP_EXECNONE,
            STOP.STOP_ZEROPAGE,
            STOP.STOP_NOSTART,
            STOP.STOP_SEGFAULT,
            STOP.STOP_ZERO_DIV,
            STOP.STOP_HLT,
            STOP.STOP_SYSCALL_ARM,
            STOP.STOP_X86_CPUID,
        }

    def __getstate__(self):
        parent_ret = super().__getstate__()
        return (parent_ret, self._block_details_cache, self._stop_block_addrs_cache, self._stop_reasons_to_track)

    def __setstate__(self, args):
        super().__setstate__(args[0])
        self._block_details_cache = args[1]
        self._stop_block_addrs_cache = args[2]
        self._stop_reasons_to_track = args[3]

    def __check(self, num_inst=None, **kwargs):  # pylint: disable=unused-argument
        state = self.state
        if o.UNICORN not in state.options:
            l.debug("Unicorn-engine is not enabled.")
            return False

        if uc_module is None or _UC_NATIVE is None:
            if once("unicorn_install_warning"):
                l.error(
                    "You are attempting to use unicorn engine support even though it or the angr native layer "
                    "isn't installed"
                )
            return False

        self.__countdown(state)

        # should the countdown still be updated if we're not stepping a whole block?
        # current decision: leave it updated, since we are moving forward
        if num_inst is not None:
            # we don't support single stepping with unicorn
            return False

        unicorn = state.unicorn  # shorthand

        # if we have a concrete target we want the program to synchronize the segment
        # registers before, otherwise undefined behavior could happen.
        if state.project.concrete_target and self.project.arch.name in ("x86", "x86_64"):
            if not state.concrete.segment_registers_initialized:
                l.debug("segment register must be synchronized with the concrete target before using unicorn engine")
                return False
        if state.regs.ip.symbolic:
            l.debug("symbolic IP!")
            return False
        if unicorn.countdown_symbolic_stop > 0:
            l.info("not enough blocks since symbolic stop (%d more)", unicorn.countdown_symbolic_stop)
            return False
        if unicorn.countdown_unsupported_stop > 0:
            l.info(
                "not enough blocks since unsupported VEX statement/expression stop (%d more)",
                unicorn.countdown_unsupported_stop,
            )
            return False
        if unicorn.countdown_nonunicorn_blocks > 0:
            l.info("not enough runs since last unicorn (%d)", unicorn.countdown_nonunicorn_blocks)
            return False
        if unicorn.countdown_stop_point > 0:
            l.info("not enough blocks since stop point (%d more)", unicorn.countdown_stop_point)
        elif o.UNICORN_SYM_REGS_SUPPORT not in state.options and not unicorn._check_registers():
            l.info("failed register check")
            return False

        if state.addr in self._stop_block_addrs_cache:
            l.info("Block will likely not execute in native interface")
            return False

        return True

    @staticmethod
    def __countdown(state):
        state.unicorn.countdown_nonunicorn_blocks -= 1
        state.unicorn.countdown_symbolic_stop -= 1
        state.unicorn.countdown_unsupported_stop -= 1
        state.unicorn.countdown_stop_point -= 1

    def _execute_block_instrs_in_vex(self, block_details):
        if block_details["block_addr"] not in self._block_details_cache:
            vex_block = self._get_vex_block_details(block_details["block_addr"], block_details["block_size"])
            self._block_details_cache[block_details["block_addr"]] = vex_block
        else:
            vex_block = self._block_details_cache[block_details["block_addr"]]

        # Save breakpoints for restoring later
        saved_mem_read_breakpoints = copy.copy(self.state.inspect._breakpoints["mem_read"])
        saved_mem_write_breakpoints = copy.copy(self.state.inspect._breakpoints["mem_write"])
        for reg_name, reg_value in block_details["registers"]:
            self.state.registers.store(reg_name, reg_value, inspect=False, disable_actions=True)

        self.state.scratch.set_tyenv(vex_block.tyenv)
        for stmt_entry in block_details["stmts"]:
            self._instr_mem_reads = list(stmt_entry["mem_dep"])  # pylint:disable=attribute-defined-outside-init
            if self._instr_mem_reads:
                # Insert breakpoint to set the correct memory read address
                self.state.inspect.b("mem_read", when=BP_BEFORE, action=self._set_correct_mem_read_addr)

            self.state.inspect.b("mem_write", when=BP_AFTER, action=self._save_mem_write_addrs)
            execute_default_exit = True
            # Execute handler from HeavyVEXMixin for the statement
            vex_stmt = vex_block.statements[stmt_entry["stmt_idx"]]
            self.stmt_idx = stmt_entry["stmt_idx"]  # pylint:disable=attribute-defined-outside-init
            try:
                super()._handle_vex_stmt(vex_stmt)  # pylint:disable=no-member
            except VEXEarlyExit:
                # Only one path is satisfiable in this branch.
                execute_default_exit = False

            # Restore breakpoints
            self.state.inspect._breakpoints["mem_read"] = copy.copy(saved_mem_read_breakpoints)
            self.state.inspect._breakpoints["mem_write"] = copy.copy(saved_mem_write_breakpoints)
            del self._instr_mem_reads

        if execute_default_exit and block_details["has_symbolic_exit"]:
            # Process block's default exit
            self.stmt_idx = DEFAULT_STATEMENT  # pylint:disable=attribute-defined-outside-init
            super()._handle_vex_defaultexit(vex_block.next, vex_block.jumpkind)  # pylint:disable=no-member

        # Restore breakpoints
        for succ_state in self.successors.successors:
            succ_state.inspect._breakpoints["mem_read"] = copy.copy(saved_mem_read_breakpoints)
            succ_state.inspect._breakpoints["mem_write"] = copy.copy(saved_mem_read_breakpoints)

        del self.stmt_idx

    def _execute_symbolic_instrs(self, syscall_data):
        recent_bbl_addrs = None
        stop_details = None

        self._instr_mem_write_addrs = set()  # pylint:disable=attribute-defined-outside-init
        for block_details in self.state.unicorn._get_details_of_blocks_with_symbolic_vex_stmts():
            self.state.scratch.guard = self.state.solver.true
            try:
                if self.state.os_name == "CGC" and block_details["block_addr"] in {
                    self.state.unicorn.cgc_random_addr,
                    self.state.unicorn.cgc_receive_addr,
                }:
                    # Re-execute CGC syscall
                    reg_vals = dict(block_details["registers"])
                    curr_regs = self.state.regs
                    # If any regs are not present in the block details for re-execute, they are probably symbolic and so
                    # were not saved in native interface. Use current register values in those cases: they should have
                    # correct values right now.
                    if block_details["block_addr"] == self.state.unicorn.cgc_receive_addr:
                        # rx_bytes argument is set to 0 since we care about updating symbolic values only
                        syscall_args = [
                            reg_vals.get("ebx", curr_regs.ebx),
                            reg_vals.get("ecx", curr_regs.ecx),
                            reg_vals.get("edx", curr_regs.edx),
                            0,
                        ]
                        syscall_simproc = self.state.project.simos.syscall_from_number(3, abi=None)
                        syscall_simproc.arch = self.state.arch
                        syscall_simproc.project = self.state.project
                        syscall_simproc.state = self.state
                        syscall_simproc.cc = self.state.project.simos.syscall_cc(self.state)
                        ret_val = getattr(syscall_simproc, syscall_simproc.run_func)(*syscall_args)
                        self.state.registers.store("eax", ret_val, inspect=False, disable_actions=True)
                    elif block_details["block_addr"] == self.state.unicorn.cgc_random_addr:
                        syscall_simproc = self.state.project.simos.syscall_from_number(7, abi=None)
                        # rnd_bytes argument is set to 0 since we care about updating symbolic values only
                        syscall_args = [reg_vals.get("ebx", curr_regs.ebx), reg_vals.get("ecx", curr_regs.ecx), 0]
                        if o.UNICORN_HANDLE_CGC_RANDOM_SYSCALL in self.state.options:
                            # Update concrete value before invoking syscall
                            concrete_data = b""
                            curr_size = 0
                            max_size = self.state.solver.eval(syscall_args[1])
                            while curr_size != max_size:
                                next_entry = syscall_data["random"].pop(0)
                                curr_size = curr_size + next_entry[1]
                                endianness = "little" if self.state.arch.memory_endness == "Iend_LE" else "big"
                                concrete_data = concrete_data + next_entry[0].to_bytes(next_entry[1], endianness)
                        else:
                            concrete_data = None

                        syscall_simproc.arch = self.state.arch
                        syscall_simproc.project = self.state.project
                        syscall_simproc.state = self.state
                        syscall_simproc.cc = self.state.project.simos.syscall_cc(self.state)
                        ret_val = getattr(syscall_simproc, syscall_simproc.run_func)(*syscall_args, concrete_data)
                        self.state.registers.store("eax", ret_val, inspect=False, disable_actions=True)
                else:
                    if block_details["has_symbolic_exit"]:
                        curr_succs_count = len(self.successors.successors)
                        if not recent_bbl_addrs:
                            recent_bbl_addrs = self.state.unicorn.get_recent_bbl_addrs()

                        if not stop_details:
                            stop_details = self.state.unicorn.get_stop_details()

                    self._execute_block_instrs_in_vex(block_details)
                    if block_details["has_symbolic_exit"]:
                        curr_succs = self.successors.successors
                        if len(curr_succs) == curr_succs_count + 1:
                            # There is only one newly added satisfiable successor state and so that is the state that
                            # follows path being traced
                            self.state = curr_succs[curr_succs_count]
                            self.successors.flat_successors.remove(self.state)
                            self.successors.all_successors.remove(self.state)
                            self.successors.successors.remove(self.state)
                        else:
                            # There are multiple satisfiable states. Use the state's record of basic blocks executed
                            # and block where native interface stopped to determine which state followed the path traced
                            # till now

                            next_block_on_path = None
                            if block_details["block_hist_ind"] + 1 < len(recent_bbl_addrs):
                                next_block_on_path = recent_bbl_addrs[block_details["block_hist_ind"] + 1]
                            else:
                                next_block_on_path = stop_details.block_addr

                            for succ in curr_succs[curr_succs_count:]:
                                if succ.addr == next_block_on_path:
                                    self.state = succ
                                    self.successors.flat_successors.remove(succ)
                                    self.successors.successors.remove(succ)
                                    break
                            else:
                                raise Exception("Multiple valid successor states found but none followed the trace!")
            except SimValueError as e:
                l.error(e)

        del self._instr_mem_write_addrs

    def _get_vex_block_details(self, block_addr, block_size):
        # Mostly based on the lifting code in HeavyVEXMixin
        # pylint:disable=no-member
        irsb = super().lift_vex(addr=block_addr, state=self.state, size=block_size)
        if irsb.size == 0:
            if irsb.jumpkind == "Ijk_NoDecode":
                if not self.state.project.is_hooked(irsb.addr):
                    raise SimIRSBNoDecodeError(
                        f"IR decoding error at 0x{irsb.addr:02x}. You can hook this instruction"
                        " with a python replacement using project.hook"
                        f"(0x{irsb.addr:02x}, your_function, length=length_of_instruction)."
                    )

                raise SimIRSBError("Block is hooked with custom code but original block was executed in unicorn")

            raise SimIRSBError(f"Empty IRSB found at 0x{irsb.addr:02x}.")

        return irsb

    def _set_correct_mem_read_addr(self, state):
        assert len(self._instr_mem_reads) != 0
        mem_read_val = b""
        mem_read_size = 0
        mem_read_address = None
        mem_read_taint_map = []
        while mem_read_size != state.inspect.mem_read_length and self._instr_mem_reads:
            next_val = self._instr_mem_reads.pop(0)
            if not mem_read_address:
                mem_read_address = next_val["address"]

            if next_val["symbolic"]:
                if next_val["address"] in self._instr_mem_write_addrs:
                    # This address was modified during re-execution. Ignore taint reported by native interface
                    mem_read_taint_map.append(-1)
                else:
                    mem_read_taint_map.append(1)
            else:
                mem_read_taint_map.append(0)

            mem_read_size += 1
            mem_read_val += next_val["value"]

        assert state.inspect.mem_read_length == mem_read_size
        state.inspect.mem_read_address = state.solver.BVV(mem_read_address, state.inspect.mem_read_address.size())
        if mem_read_taint_map.count(-1) != mem_read_size:
            # Since read is might need bitmap adjustment, insert breakpoint to return the correct concrete value
            self.state.inspect.b(
                "mem_read",
                when=BP_AFTER,
                action=functools.partial(
                    self._set_correct_mem_read_val, value=mem_read_val, taint_map=mem_read_taint_map
                ),
            )

    def _set_correct_mem_read_val(self, state, value, taint_map):  # pylint: disable=no-self-use
        state.inspect._breakpoints["mem_read"].pop()
        if taint_map.count(0) == state.inspect.mem_read_length:
            # The value is completely concrete
            if state.arch.memory_endness == archinfo.Endness.LE:
                state.inspect.mem_read_expr = state.solver.BVV(value[::-1])
            else:
                state.inspect.mem_read_expr = state.solver.BVV(value)
        else:
            # The value may be partially concrete. Set the symbolic bitmap to read correct value and restore it
            mem_read_addr = state.solver.eval(state.inspect.mem_read_address)
            mem_read_len = state.inspect.mem_read_length
            saved_taints = []
            for offset in range(mem_read_len):
                page_num, page_off = state.memory._divide_addr(mem_read_addr + offset)
                page_obj = state.memory._get_page(page_num, writing=False)
                saved_taints.append(page_obj.symbolic_bitmap[page_off])

            restore_taints = False
            if saved_taints != taint_map:
                # Symbolic bitmap needs fixing before reading value from memory.
                restore_taints = True
                for offset, expected_taint in enumerate(taint_map):
                    if expected_taint != -1:
                        page_num, page_off = state.memory._divide_addr(mem_read_addr + offset)
                        page_obj = state.memory._get_page(page_num, writing=False)
                        page_obj.symbolic_bitmap[page_off] = expected_taint

            curr_value = state.memory.load(
                mem_read_addr, mem_read_len, endness=state.arch.memory_endness, inspect=False, disable_actions=True
            )
            if restore_taints:
                for offset, saved_taint in enumerate(saved_taints):
                    page_num, page_off = state.memory._divide_addr(mem_read_addr + offset)
                    page_obj = state.memory._get_page(page_num, writing=False)
                    page_obj.symbolic_bitmap[page_off] = saved_taint

            if taint_map.count(0) != 0:
                # Update concrete bytes using values reported by native interface
                curr_value_bytes = curr_value.chop(8)
                if state.arch.memory_endness == archinfo.Endness.LE:
                    curr_value_bytes.reverse()

                for offset, expected_taint in enumerate(taint_map):
                    if expected_taint == 0:
                        curr_value_bytes[offset] = state.solver.BVV(value[offset], 8)

                if state.arch.memory_endness == archinfo.Endness.LE:
                    curr_value_bytes = reversed(curr_value_bytes)

                curr_value = state.solver.Concat(*curr_value_bytes)

            state.inspect.mem_read_expr = curr_value

    def _save_mem_write_addrs(self, state):
        mem_write_addr = state.solver.eval(state.inspect.mem_write_address)
        self._instr_mem_write_addrs.update(range(mem_write_addr, mem_write_addr + state.inspect.mem_write_length))

    def process_successors(self, successors, **kwargs):
        state = self.state
        if not self.__check(**kwargs):
            return super().process_successors(successors, **kwargs)

        extra_stop_points = kwargs.get("extra_stop_points", None)
        last_block_details = kwargs.get("last_block_details", None)
        step = kwargs.get("step", None)
        if extra_stop_points is None:
            extra_stop_points = set(self.project._sim_procedures)
        else:
            # convert extra_stop_points to a set
            extra_stop_points = set(extra_stop_points)
            extra_stop_points.update(self.project._sim_procedures)
        if successors.addr in extra_stop_points:
            # trying to start unicorn execution on a stop point, fallback to next engine
            return super().process_successors(successors, **kwargs)

        successors.sort = "Unicorn"

        # add all instruction breakpoints as extra_stop_points
        if state.supports_inspect:
            for bp in state.inspect._breakpoints["instruction"]:
                # if there is an instruction breakpoint on every instruction, it does not make sense
                # to use unicorn.
                if "instruction" not in bp.kwargs:
                    l.info("disabling unicorn because of breakpoint on every instruction")
                    return super().process_successors(successors, **kwargs)

                # add the breakpoint to extra_stop_points. We don't care if the breakpoint is BP_BEFORE or
                # BP_AFTER, this is only to stop unicorn when we get near a breakpoint. The breakpoint itself
                # will then be handled by another engine that can more accurately step instruction-by-instruction.
                extra_stop_points.add(bp.kwargs["instruction"])

        # initialize unicorn plugin
        try:
            syscall_data = kwargs["syscall_data"] if "syscall_data" in kwargs else None
            fd_bytes = kwargs["fd_bytes"] if "fd_bytes" in kwargs else None
            state.unicorn.setup(syscall_data=syscall_data, fd_bytes=fd_bytes)
        except SimValueError:
            # it's trying to set a symbolic register somehow
            # fail out, force fallback to next engine
            return super().process_successors(successors, **kwargs)

        try:
            state.unicorn.set_stops(extra_stop_points)
            if last_block_details is not None:
                state.unicorn.set_last_block_details(last_block_details)
            state.unicorn.set_tracking(
                track_bbls=o.UNICORN_TRACK_BBL_ADDRS in state.options,
                track_stack=o.UNICORN_TRACK_STACK_POINTERS in state.options,
            )
            state.unicorn.hook()
            state.unicorn.start(step=step)
            self._execute_symbolic_instrs(syscall_data=syscall_data)
            state.unicorn.finish(self.state)
        finally:
            state.unicorn.destroy(self.state)

        state = self.state
        if state.unicorn.stop_reason in self._stop_reasons_to_track:
            if state.unicorn.steps == 0:
                self._stop_block_addrs_cache.add(state.addr)
            else:
                self._stop_block_addrs_cache.add(state.unicorn.stop_details.block_addr)

        if state.unicorn.steps == 0 or state.unicorn.stop_reason == STOP.STOP_NOSTART:
            # fail out, force fallback to next engine
            # TODO: idk what the consequences of this might be. If this failed step can actually change non-unicorn
            # state then this is bad news.
            return super().process_successors(successors, **kwargs)

        description = f"Unicorn ({STOP.name_stop(state.unicorn.stop_reason)} after {state.unicorn.steps} steps)"

        state.history.recent_block_count += state.unicorn.steps
        state.history.recent_description = description

        # this can be expensive, so check first
        if state.supports_inspect:
            for bp in state.inspect._breakpoints["irsb"]:
                if bp.check(state, BP_AFTER):
                    for bbl_addr in state.history.recent_bbl_addrs:
                        state._inspect("irsb", BP_AFTER, address=bbl_addr)
                    break

        if state.unicorn.stop_reason in (
            STOP.symbolic_stop_reasons | STOP.unsupported_reasons
        ) or state.unicorn.stop_reason in (STOP.STOP_UNKNOWN_MEMORY_WRITE_SIZE, STOP.STOP_VEX_LIFT_FAILED):
            l.info(state.unicorn.stop_message)

        if state.unicorn.jumpkind.startswith("Ijk_Sys"):
            state.ip = state.unicorn._syscall_pc
        successors.add_successor(state, state.ip, state.solver.true, state.unicorn.jumpkind)

        successors.description = description
        successors.processed = True

        return None



# pylint: disable=missing-class-docstring
import glob
import importlib
import os
import platform
import shutil
import subprocess
import sys
from distutils.command.build import build as st_build
from distutils.util import get_platform

import pkg_resources
from setuptools import Command, setup
from setuptools.command.develop import develop as st_develop
from setuptools.errors import LibError

if sys.platform == "darwin":
    library_file = "angr_native.dylib"
elif sys.platform in ("win32", "cygwin"):
    library_file = "angr_native.dll"
else:
    library_file = "angr_native.so"


def _build_native():
    try:
        importlib.import_module("pyvex")
    except ImportError as e:
        raise LibError("You must install pyvex before building angr") from e

    try:
        importlib.import_module("unicorn")
    except ImportError as e:
        raise LibError("You must install unicorn before building angr") from e

    env = os.environ.copy()
    env_data = (
        ("UNICORN_INCLUDE_PATH", "unicorn", "include"),
        ("UNICORN_LIB_PATH", "unicorn", "lib"),
        ("UNICORN_LIB_FILE", "unicorn", "lib\\unicorn.lib"),
        ("PYVEX_INCLUDE_PATH", "pyvex", "include"),
        ("PYVEX_LIB_PATH", "pyvex", "lib"),
        ("PYVEX_LIB_FILE", "pyvex", "lib\\pyvex.lib"),
    )
    for var, pkg, fnm in env_data:
        try:
            env[var] = pkg_resources.resource_filename(pkg, fnm)
        except KeyError:
            pass

    if sys.platform == "win32":
        cmd = ["nmake", "/f", "Makefile-win"]
    elif shutil.which("gmake") is not None:
        cmd = ["gmake"]
    else:
        cmd = ["make"]
    try:
        subprocess.run(cmd, cwd="native", env=env, check=True)
    except FileNotFoundError:
        raise LibError("Couldn't find " + cmd[0] + " in PATH")
    except subprocess.CalledProcessError as err:
        raise LibError("Error while building angr_native: " + str(err))

    shutil.rmtree("angr/lib", ignore_errors=True)
    os.mkdir("angr/lib")
    shutil.copy(os.path.join("native", library_file), "angr/lib")


def _clean_native():
    oglob = glob.glob("native/*.o")
    oglob += glob.glob("native/*.obj")
    oglob += glob.glob("native/*.so")
    oglob += glob.glob("native/*.dll")
    oglob += glob.glob("native/*.dylib")
    for fname in oglob:
        os.unlink(fname)


class build(st_build):
    def run(self, *args):
        self.execute(_build_native, (), msg="Building angr_native")
        super().run(*args)


class clean_native(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        self.execute(_clean_native, (), msg="Cleaning angr_native")


class develop(st_develop):
    def run(self):
        self.run_command("build")
        super().run()


cmdclass = {
    "build": build,
    "clean_native": clean_native,
    "develop": develop,
}


try:
    from setuptools.command.editable_wheel import editable_wheel as st_editable_wheel

    class editable_wheel(st_editable_wheel):
        def run(self):
            self.run_command("build")
            super().run()

    cmdclass["editable_wheel"] = editable_wheel
except ModuleNotFoundError:
    pass


if "bdist_wheel" in sys.argv and "--plat-name" not in sys.argv:
    sys.argv.append("--plat-name")
    name = get_platform()
    if "linux" in name:
        sys.argv.append("manylinux2014_" + platform.machine())
    else:
        # https://www.python.org/dev/peps/pep-0425/
        sys.argv.append(name.replace(".", "_").replace("-", "_"))

setup(cmdclass=cmdclass)



from typing import Set, List, Tuple, Optional, TYPE_CHECKING

import networkx

# FIXME: Remove the dependency on pyformlang
from pyformlang.finite_automaton import Epsilon, EpsilonNFA, State, Symbol

from angr.errors import AngrError
from .typevars import BaseLabel, Subtype
from .variance import Variance

if TYPE_CHECKING:
    from pyformlang.finite_automaton import DeterministicFiniteAutomaton


START_STATE = State("START")
END_STATE = State("END")


class EmptyEpsilonNFAError(AngrError):
    """
    A notification exception generated when the epsilon NFA is empty.
    """


class DFAConstraintSolver:
    """
    Implements a DFA-based graph solver.
    """

    @staticmethod
    def graph_to_epsilon_nfa(graph: networkx.DiGraph, starts: Set, ends: Set) -> EpsilonNFA:
        enfa = EpsilonNFA()

        # print("Converting graph to eNFA")

        for src, dst, data in graph.edges(data=True):
            if not data:
                symbol = Epsilon()
            else:
                assert "label" in data
                label, kind = data["label"]
                symbol = Symbol((label, kind))

            # print(src, "-----", symbol, "----->", dst)
            enfa.add_transition(State(src), symbol, State(dst))

        enfa.add_start_state(START_STATE)

        for start in starts:
            # print("Start transition", START_STATE, "----->", start)
            enfa.add_transition(START_STATE, Symbol(start), State(start))

        enfa.add_final_state(END_STATE)
        for end in ends:
            # print("End transition", end, "----->", END_STATE)
            enfa.add_transition(State(end), Symbol(end), END_STATE)

        # networkx.drawing.nx_pydot.write_dot(graph, "d:/enfa.dot")

        if enfa.is_empty():
            raise EmptyEpsilonNFAError()
        return enfa

    def generate_constraints_between(self, graph: networkx.DiGraph, starts: Set, ends: Set) -> Set:
        epsilon_nfa = self.graph_to_epsilon_nfa(graph, starts, ends)
        min_dfa: "DeterministicFiniteAutomaton" = epsilon_nfa.minimize()
        dfa_graph: networkx.MultiDiGraph = min_dfa.to_networkx()

        constraints = set()

        for final_state in min_dfa.final_states:
            for path in networkx.all_simple_edge_paths(dfa_graph, min_dfa.start_state, final_state):
                path_labels = []
                for src, dst, index in path:
                    d = dfa_graph.get_edge_data(src, dst)[index]
                    path_labels.append(d["label"])

                start_node = path_labels[0]
                end_node = path_labels[-1]

                constraint = self._check_constraint(start_node, end_node, path_labels[1:-1])
                if constraint is not None:
                    constraints.add(constraint)

        return constraints

    @staticmethod
    def _check_constraint(src, dst, string: List[Tuple[BaseLabel, str]]) -> Optional[Subtype]:
        forgets = []
        recalls = []
        for label, kind in string:
            if kind == "forget":
                forgets.append(label)
            elif kind == "recall":
                recalls.append(label)

        lhs = src
        rhs = dst
        for recall in recalls:
            lhs = lhs.recall(recall)
        for forget in reversed(forgets):
            rhs = rhs.recall(forget)

        if lhs.variance == Variance.COVARIANT and rhs.variance == Variance.COVARIANT:
            if lhs.typevar != rhs.typevar:
                return Subtype(lhs.typevar, rhs.typevar)
        return None



import operator

from archinfo.arch_soot import SootNullConstant
import claripy

from ..values import SimSootValue_StringRef, SimSootValue_ThisRef
from .base import SimSootExpr


class SimSootExpr_Condition(SimSootExpr):
    def _execute(self):
        v1 = self._translate_expr(self.expr.value1)
        v2 = self._translate_expr(self.expr.value2)
        operator_func = SimSootExpr_Condition.condition_str_to_function[self.expr.op]
        if isinstance(v1.expr, SimSootValue_ThisRef) and isinstance(v2.expr, SootNullConstant) and v1.expr.symbolic:
            # TODO fix this
            self.expr = operator_func(claripy.BVS("obf_ref", 32), claripy.BVV(0, 32))
        elif isinstance(v1.expr, (SootNullConstant, SimSootValue_StringRef)) or isinstance(
            v2.expr, (SootNullConstant, SimSootValue_StringRef)
        ):
            self.expr = claripy.true if operator_func(v1.expr, v2.expr) else claripy.false
        else:
            self.expr = operator_func(v1.expr, v2.expr)

    # Java only has signed primitives
    condition_str_to_function = {
        "eq": operator.eq,
        "ne": operator.ne,
        "ge": claripy.SGE,
        "gt": claripy.SGT,
        "le": claripy.SLE,
        "lt": claripy.SLT,
        # TODO others...
    }



from typing import Optional, Set, Dict, Tuple, Any, TYPE_CHECKING

from .clinic import Clinic
from .structured_codegen import BaseStructuredCodeGenerator

if TYPE_CHECKING:
    from angr.analyses.decompiler.optimization_passes.expr_op_swapper import OpDescriptor


class DecompilationCache:
    """
    Caches key data structures that can be used later for refining decompilation results, such as retyping variables.
    """

    __slots__ = (
        "addr",
        "type_constraints",
        "var_to_typevar",
        "codegen",
        "clinic",
        "ite_exprs",
        "binop_operators",
    )

    def __init__(self, addr):
        self.addr = addr
        self.type_constraints: Optional[Set] = None
        self.var_to_typevar: Optional[Dict] = None
        self.codegen: Optional[BaseStructuredCodeGenerator] = None
        self.clinic: Optional[Clinic] = None
        self.ite_exprs: Optional[Set[Tuple[int, Any]]] = None
        self.binop_operators: Optional[Dict["OpDescriptor", str]] = None

    @property
    def local_types(self):
        return self.clinic.variable_kb.variables[self.addr].types



from . import SimConcretizationStrategy


class SimConcretizationStrategyNonzeroRange(SimConcretizationStrategy):
    """
    Concretization strategy that resolves a range in a non-zero location.
    """

    def __init__(self, limit, **kwargs):
        super().__init__(**kwargs)
        self._limit = limit

    def _concretize(self, memory, addr, extra_constraints=None, **kwargs):
        mn, mx = self._range(memory, addr, extra_constraints=extra_constraints, **kwargs)
        if mx - mn <= self._limit:
            child_constaints = (addr != 0,)
            if extra_constraints is not None:
                child_constaints += tuple(extra_constraints)
            return self._eval(memory, addr, self._limit, extra_constraints=child_constaints, **kwargs)
        return None
